
using Godot;
using GArray = Godot.Collections.Array;
using System.Collections.Concurrent;

// cube data struct
public struct Cube {
  public int Id;
  public Transform Position;
  public Vector3 Size;
  public GArray MeshArray;
  public RID Mesh;
  public RID Instance;
  public RID Body;
  public RID Shape;
}

// cubes factory
public static class Cubes {
  // "thread safe" dictionary
  public static ConcurrentDictionary<int, Cube> Lookup = new ConcurrentDictionary<int, Cube>();

  // for easier access, populated
  // on scene root script _Ready
  public static RID Scenario;
  public static RID Space;

  // reuse material for simplicity
  static SpatialMaterial material;

  public static void Add(int id, Vector3 size, Transform position) {
    if (!Lookup.TryGetValue(id, out _)) {
      if (material == null) {
        material = new SpatialMaterial();
        material.AlbedoColor = new Color("#ec6161");
      }

      Cube cube = new Cube();
      cube.Id = id;
      cube.Size = size;
      cube.Position = position;

      GArray arrays = GenerateMesh(size);
      RID mesh = VisualServer.MeshCreate();
      VisualServer.MeshAddSurfaceFromArrays(mesh, VisualServer.PrimitiveType.Triangles, arrays);
      VisualServer.MeshSurfaceSetMaterial(mesh, 0, material.GetRid());

      RID instance = VisualServer.InstanceCreate();
      VisualServer.InstanceSetTransform(instance, position);
      VisualServer.InstanceSetScenario(instance, Scenario);
      VisualServer.InstanceSetBase(instance, mesh);

      // physics
      // RID body = PhysicsServer.BodyCreate();
      // PhysicsServer.BodySetMode(body, PhysicsServer.BodyMode.Static);
      // RID shape = PhysicsServer.ShapeCreate(PhysicsServer.ShapeType.Box);
      // PhysicsServer.ShapeSetData(shape, new Vector3(cube.Size.x / 2f, cube.Size.y / 2f, cube.Size.z / 2f));
      // PhysicsServer.BodyAddShape(body, shape, new Transform(Basis.Identity, new Vector3(0f, cube.Size.y / 2f, 0f)));
      // PhysicsServer.BodySetSpace(body, Space);
      // PhysicsServer.BodySetState(body, PhysicsServer.BodyState.Transform, cube.Position);

      cube.MeshArray = arrays;
      cube.Mesh = mesh;
      cube.Instance = instance;
      // cube.Body = body;
      // cube.Shape = shape;

      Lookup.TryAdd(id, cube);
    }
  }

  public static void Remove(int id) {
    if (Lookup.TryGetValue(id, out Cube cube)) {
      VisualServer.FreeRid(cube.Instance);
      // PhysicsServer.FreeRid(cube.Shape);
      // PhysicsServer.FreeRid(cube.Body);

      Lookup.TryRemove(id, out _);
    }
  }

  public static GArray GenerateMesh(Vector3 size) {
    float halfX = size.x / 2f;
    float halfZ = size.z / 2f;

    Vector3[] vertices = new Vector3[] {
      // top face
      new Vector3(0 - halfX, size.y, 0 - halfZ),
      new Vector3(0 + halfX, size.y, 0 - halfZ),
      new Vector3(0 + halfX, size.y, 0 + halfZ),
      new Vector3(0 - halfX, size.y, 0 + halfZ),

      // forward face
      new Vector3(0 - halfX, 0, 0 - halfZ),
      new Vector3(0 + halfX, 0, 0 - halfZ),
      new Vector3(0 + halfX, size.y, 0 - halfZ),
      new Vector3(0 - halfX, size.y, 0 - halfZ),

      // back face
      new Vector3(0 + halfX, 0, 0 + halfZ),
      new Vector3(0 - halfX, 0, 0 + halfZ),
      new Vector3(0 - halfX, size.y, 0 + halfZ),
      new Vector3(0 + halfX, size.y, 0 + halfZ),

      // right face
      new Vector3(0 + halfX, 0, 0 - halfZ),
      new Vector3(0 + halfX, 0, 0 + halfZ),
      new Vector3(0 + halfX, size.y, 0 + halfZ),
      new Vector3(0 + halfX, size.y, 0 - halfZ),

      // left face
      new Vector3(0 - halfX, 0, 0 + halfZ),
      new Vector3(0 - halfX, 0, 0 - halfZ),
      new Vector3(0 - halfX, size.y, 0 - halfZ),
      new Vector3(0 - halfX, size.y, 0 + halfZ)
    };

    int[] indices = new int[] {
      // roof
      0, 1, 2, 
      0, 2, 3,
      // -z side
      4, 5, 6,
      4, 6, 7,
      // z side
      8, 9, 10,
      8, 10, 11,
      // x side
      12, 13, 14,
      12, 14, 15,
      // -x side
      16, 17, 18,
      16, 18, 19
    };

    Vector3[] normals = new Vector3[] {
      Vector3.Up, Vector3.Up, Vector3.Up, Vector3.Up,
      Vector3.Forward, Vector3.Forward, Vector3.Forward, Vector3.Forward,
      Vector3.Back, Vector3.Back, Vector3.Back, Vector3.Back,
      Vector3.Right, Vector3.Right, Vector3.Right, Vector3.Right,
      Vector3.Left, Vector3.Left, Vector3.Left, Vector3.Left
    };

    GArray arrays = new GArray();
    arrays.Resize((int)Mesh.ArrayType.Max);

    arrays[(int)Mesh.ArrayType.Vertex] = vertices;
    arrays[(int)Mesh.ArrayType.Index] = indices;
    arrays[(int)Mesh.ArrayType.Normal] = normals;

    return arrays;
  }
}